package Petal::Application;

use strict;
use warnings;

use CGI::Carp;
use File::Find;
use Data::Dumper;
use Apache::Constants qw(:response);
use Petal::Application::Template;
use Petal::Application::TemplateHelper;

my $debug = 0;

sub new {
	my ($class, $r) = @_;
	my $self = {};
	bless $self, $class;

	# variable accessors
	for my $accessor qw(
			name
			base_url
			request 
			template
			template_path 
			template_file
			default_path
			response
			location
			current_module
			current_hook
			direct_output
		) {
		$self->new_accessor($accessor);
	}
	$self->request($r);

	my $template_dir = $self->find_template_dir();
	my $module_dir = $self->find_module_dir();

	$self->template_path($template_dir);

	find({ no_chdir => 1, wanted => sub {
		s/^$module_dir(\/*)//;
		return unless /\.pm$/;

		my $fn = $_;

		my $package_name = join '::', split /\//, $fn;
		$package_name =~ s/\.pm$//;
		my $unique_namespace = $self->unique_namespace();
		$package_name = $unique_namespace . '::Module::' . $package_name;

		# require filename
		my $module_fn = $module_dir . '/' . $fn;

		open my($fh), $module_fn;
		local $/ = undef;
		my $module_text = <$fh>;
		close $fh;

		$module_text =~ s/package Module\:\:\S+/package $package_name;/m;
		$module_text = "#line 1 $fn\n" . $module_text;

		eval $module_text;
		if($@) {
			carp("File $_ reported errors: $@. Skipped.");
			return;
		}

		# instantiate object
		my $o = $package_name->new($self);
		$self->{modules}->{$package_name} = $o;
		carp("Module $_ loaded.") if $debug;
	} }, ($module_dir));

	return $self;
}

sub unique_namespace {
	my ($self) = @_;
	my $r = $self->request;

	my $location = $r->location;
	$location ||= "Root";
	$location =~ s/^\/?(.*)\/?$/$1/;

	return 'Petal::Application::ModuleNameSpace::' . 
		join('::', split(/\//, $location));
}

sub find_template_dir {
	my ($self) = @_;
	my $r = $self->request;

	my $template_dir = $r->dir_config('PETAL_APPLICATION_TEMPLATE_PATH');
	return $template_dir if $template_dir;

	return $r->document_root . $r->location . 'usr/templates';
}

sub find_module_dir {
	my ($self) = @_;
	my $r = $self->request;

	my $module_dir = $r->dir_config('PETAL_APPLICATION_MODULE_PATH');
	return $module_dir if $module_dir;

	return $r->document_root . $r->location . 'usr/modules';
}

sub new_accessor {
	my ($self, $varname) = @_;
	return if $self->{accessors}->{$varname};
	no strict 'refs'; # have to turn them off for this one
	*$varname = sub {
		my ($self, $value) = @_;
		$self->{$varname} = $value if defined $value;
		return $self->{$varname};
	};
	$self->{accessors}->{$varname} = 1;
}

sub modules {
	my ($self) =  @_;
	return [ keys %{$self->{modules}} ];
}

sub hook {
	my ($self, $modules, $hook, @args) = @_;
	$modules = [ $modules ] unless ref $modules;

	undef $self->{weights};

	# step one -- build weighted list of modules (if it doesn't already exist for this hook)
	for my $module (@$modules) {
		next if defined $self->{weights}->{$hook}->{$module};
		my $weight = 0;
		if($self->{modules}->{$module}->can('weight')) {
			$weight = $self->{modules}->{$module}->weight($hook) || 0;
		}

		$self->{weights}->{$hook}->{$module} = $weight if $self->{modules}->{$module}->can($hook);
	}

	# build $weighted_module list;
	my @weighted_modules = sort {
		$self->{weights}->{$hook}->{$a} <=> $self->{weights}->{$hook}->{$b};
	} keys %{$self->{weights}->{$hook}};

	# step two -- execute hook
	
	my $results = {};
	for my $module (@weighted_modules) {
		$results->{$module} = $self->{modules}->{$module}->$hook(@args);
	}
	return $results;
}

sub handler($$) {
	my ($class, $r) = @_;

	# get self from global variables if available
	my $self = $Petal::Application::ServerObjects{$r->location};
	if(!$self) {
		$self = $class->new($r);
		$Petal::Application::ServerObjects{$r->location} = $self;
	}
	$self->request($r);
	$self->base_url($r->location);

	return DECLINED if -f $r->filename(); # shortcut processing if file exists
	$self->response(OK); # default response

	my $base_url = $self->base_url;
	my $uri = $r->uri;
	$uri =~ /^$base_url(.*)/;
	$self->location(lc $1);

	# hook init on all modules
	carp("handler: init") if $debug;
	$self->hook($self->modules, 'init', $self);

	# shortcut if this module doesn't even exist
	carp("finding url") if $debug;
	return $self->response() unless $self->find_target();

	# store GET/POST vars internally
	carp("normalizing get vars") if $debug;
	$self->set_params();

	# hook header on all modules
	carp("handler: header") if $debug;
	$self->hook($self->modules, 'header', $self);

	my $output = '';
	if($self->response == OK) {
		# create Petal object
		my $template = new Petal::Application::Template( file => $self->template_file, base_dir => $self->template_path );
		$self->template($template);

		# instantiate template helper
		my $helper = new Petal::Application::TemplateHelper($self);
		$template->param('app', $helper);

		# pre output
		carp("handler: preout") if $debug;
		$self->hook($self->modules, 'preout', $self);

		if($self->response == OK) {
			# hook page_hook on proper module
			carp("handler: page") if $debug;
			$self->direct_output(0);
			my $page = $self->page_hook();
			if($self->direct_output) {
				$output = (values %$page)[0];
				$output .= "\n"; # fixes a stupid bug with apache and chunked encoding
			}
		}

		if($self->response == OK) {
			# give modules a chance to set template variables
			carp("handler: preprocess") if $debug;
			$self->hook($self->modules, 'preprocess', $self);
		}

		# output
		if($self->response == OK) {
			$output = $template->process() unless $self->direct_output;
			carp("handler: postout") if $debug;
			$self->hook($self->modules, 'postout', $self, \$output);
		}
	}

	if ($self->response == OK) {
		$r->content_type("text/html") unless defined $r->content_type(); # default content-type
		$r->send_http_header();
		print $output;
	}

	# hook exit on all modules
	carp("handler: cleanup") if $debug;
	$self->hook($self->modules, 'cleanup', $self);

	carp("End handler()") if $debug;
	return $self->response; # we handled it.
}

sub find_target {
	my $self = shift;
	my $r = $self->request();

	my $location = $self->location;

	if(!$location) {
		$self->redirect();
		return;
	}

	my @pieces = split /\//, $location;

	my $hook = pop @pieces if @pieces > 1;

	while(@pieces) {
		# translate module name
		#  system_settings becomes SystemSettings
		my $module = $self->unique_namespace . '::Module::' . join('::', map { $_ = ucfirst; s/_(.)/uc $1/ge; $_ } @pieces);

		# set default hook to page() if necessary
		if(defined $hook) {
			$hook =~ s/\./_/g;
			$hook = 'page_' . $hook;
		}
		else {
			$hook = 'page';
		}

		# translate hook name
		#  menu.js becomes menu_js
		$hook =~ s/\./_/g;
		if(!$self->{modules}->{$module} || !$self->{modules}->{$module}->can($hook)) {
			pop @pieces if $hook eq 'page';
			$hook = undef;
			next;
		}
		else {
			$self->current_module($module);
			$self->current_hook($hook);
			return 1;
		}
	}

	$self->response(NOT_FOUND);
	return 0;
}

sub page_hook {
	my $self = shift;
	return $self->hook($self->current_module, $self->current_hook, $self);
}

sub url {
	my ($self, $path, $extra) = @_;

	$path ||= $self->default_path;
	return $path if $path =~ /^(http:|https:)/; # handle full urls

	my $url = $self->base_url . $path;
	if(defined $extra) {
		$url .= ($self->base_url =~ /\?/ ? '&' : '?') . $extra;

	}
	return $url;
}

sub set_params {
	my ($self) = @_;
	my $r = $self->request;

	$self->{get} = {};
	$self->{post} = {};
	$self->{param} = {};

	# handle GET vars
	my @get = $r->args;
	while(@get) {
		my $key = shift @get;
		last unless @get && defined $key; # bypass if the url ends with &
		my $value = shift @get;

		push @{$self->{get}->{$key}}, $value;
		push @{$self->{param}->{$key}}, $value;
	}

	# handle POST vars
	my @post = $r->content;
	while(@post) {
		my $key = shift @post;
		last unless @post && defined $key; # bypass if the url ends with &
		my $value = shift @post;

		push @{$self->{post}->{$key}}, $value;
		push @{$self->{param}->{$key}}, $value;
	}
}

sub get {
	my ($self, $param, @values) = @_;
	$self->{get}->{$param} = \@values if @values;
	if($param) {
		my $rref = $self->{get}->{$param};
		if(wantarray) {
			$rref && return @$rref;
			return ();
		}
		else {
			$rref && return $rref->[-1];
			return undef;
		}
	}
	return keys %{$self->{get}};
}

sub post {
	my ($self, $param, @values) = @_;
	$self->{post}->{$param} = \@values if @values;
	if($param) {
		my $rref = $self->{post}->{$param};
		if(wantarray) {
			$rref && return @$rref;
			return ();
		}
		else {
			$rref && return $rref->[-1];
			return undef;
		}
	}
	return keys %{$self->{post}};
}

sub param {
	my ($self, $param, @values) = @_;
	$self->{param}->{$param} = \@values if @values;
	if($param) {
		my $rref = $self->{param}->{$param};
		if(wantarray) {
			$rref && return @$rref;
			return ();
		}
		else {
			$rref && return $rref->[-1];
			return undef;
		}
	}
	return keys %{$self->{param}};
}

# copies given parameters from get/post to template
sub param2tpl {
	my ($self, @params) = @_;
	@params = $self->param() unless @params;

	for my $param (@params) {
		$self->template->param($param, $self->param($param));
	}
}

sub redirect {
	my ($self, @url) = @_;
	my $r = $self->request();

	if(! @url) {
		@url = ($self->default_path);
	}

	$r->err_header_out("Pragma", "no-cache");
	$r->header_out("Location" => $self->url(@url));
	$self->response(REDIRECT);
}

sub dbg {
	shift;
	carp Dumper \@_;
}

1;

__END__

=head1 NAME

Petal::Application - Web application framework base on L<Petal>

=head1 SYNOPSIS

	PerlModule Petal::Application

	<Location />
		SetHandler perl-script
		PerlHandler Petal::Application
	</Location
	
=head1 DESCRIPTION

Petal::Application (or P::A hereafter) is a tool for creating large 
scale business web applications quickly and efficiently.

P::A works by translating the request url into a module and page_* hook.
The module name is seen by the browser as the path to a file. For
example, given the url http://sample.com/module/submodule/new,
P::A looks in the Module::Submodule package for a method called page_new
and executes it. If Module::Submodule does not exist, P::A will try to
execute Module::Submodule::page to allow for global hooking. If there is
no Module::Submodule::page, P::A continues to Module::page. If no hooks
are found to match a request, P::A returns a 404 (Not Found) to the
user. Other hooks are executed on all modules when requested.

Note: P::A first does a check to see if the requested filename exists.
For example, if your document root looks like:

	myapp/
		images/
			image.gif
		usr/
			modules/
				User.pm
			templates/
				user/
					home.html

and a request comes in for /myapp/images/image.gif, P::A will decline to
serve the request and let apache handle it. It will not decline to
serve usr/ files. Keep this in mind, as it makes having a Usr.pm module
or an Images.pm module impossible. This is more of a feature than a
limitation. For example, If you have an Images.pm, which contains a
page() hook, and a request for /myapp/images/nonexistant.gif comes in,
your app has a chance to serve it via Images::page().

Hooks are ordered by a weighting mechanism. If a package has the method
weight, P::A executes this hook to determine in which order to executes
hooks.

=head2 Hooks

Modules interact with P::A by providing specially named methods, or
hooks. Hooks are called by P::A during various phases of serving the
request. All apps are passed a reference to the main P::A application
object (commonly $app).

=over

=item new

The new() method MUST bless your class and return it. It is the only
method that is required. Since your modules are held in the server
object throughout the lifespan of the server process, you SHOULD NOT
have any per-request functionality.
	
	sub new {
		my ($self, $class, $app) = ({}, @_);
		$self->{app} = $app;
		bless $self, $class;
		return $self;
	}

=item weight

The weight() hook MUST return a HASH reference. The reference declares the
weight of all hooks. Lower weights are executed first, and higher later.
By default, all hooks are given the weight 0. Weights are arbitrary
values, which your modules use to control execution order.

	sub weight {
		return {
			init => -100, # execute this as soon as possible
			preout => 100, # execute this later
		}
	}


=item init

The init() hook is called right after handler() is entered. Any
initialization that should be done for every page view should be done
here. GET and POST variables are not available yet.

=item header

The header() hook is called as soon as P::A is able to store GET and POST
variables. The header() hook is intended to be used to set headers for the
response. This hook is useful for doing access checking. For example,
used to redirect the client to a login page if the user is requesting a
privileged area and is not logged in.

	sub header {
		my ($self, $app) = @_;
		$app->request->content_type('text/plain'); 
			# set the Content-Type: header to text/plain
	}

=item preout

The preout() hook is the first hook in which access to the template is
available. It is called after header(), but before page_*(). It is
useful for preparing information that is available on every single page.

	sub preout {
		my ($self, $app) = @_;
		$app->template->param("page_title", "Hello, world!");
	}

=item page_*

The page_*() hooks do the bulk of the work. They prepare template
variables, parse GET/POST variables and select templates to present.

	sub page_new {
		my ($self, $app) = @_;

		$app->template->option("file", "/path/to/template.html");

		my $a = $app->param('a');
		return unless $a;

		... continue processing ...

	}

=item preprocess

The preprocess() hook is the last hook called before Petal is given the
template and parameters to process.

=item postout

The postout() hook is called after output has been generated ,but not
yet sent to the browser. The postout() hook can be used, for example, to
cleanup the output before sending to the browser.

	sub postout {
		my ($self, $app, $content) = @_;
		#content is a scalar ref to the content
		# do something with content
	}

=item cleanup

The cleanup() hook is called after output has been sent to the client.
Use this to cleanup after execution.

=back

=head2 Accessors

P::A allows modules to add accessors to itself for sharing objects
throughout the system. Yes, I know this is a poor design decision, but
it works for me. It is confusing to keep all of the accessors sorted
out, but in my apps, I only use 5 or 6 of them.

All accessors take up to one parameter. If no parameters are specified,
the accessors return their value. If one paramter is specified, the
accessors set their value to the given value.

The following accessors are available in the stock P::A:

=over

=item request

Returns the Apache::request object for the request.

=item response

Holds the constant from Apache::Constants for the return code of the
application. This is used in the redirect() method to exit prematurely
and redirect to a new page. Setting this to anything other than OK will
immediately stop processing and return that value from handler().

=item base_url

Holds the base url that is used when creating urls with the url()
method. Set to $app->request->location by default.

=item template

Holds the template object. Instead of using Petal directly, as it wasn't
designed to hold data and then process it gradually, this is an object
of type Petal::Application::Template. Petal::Application::Template
provides two methods:

=over

=item param($param, $value)

Gets or sets a parameter for use in the template. Basically stores a
local hashref where $param is the key and $value is the value. Returns
the value of $param if $value is undefined.

=item option($param, $value)

Gets or sets an option to be passed when creating the Petal object. The
most useful ones are 'file' and 'input'.

=back

=item template_path

Returns the path in which Petal expects to find the templates. Defaults
to <application root web directory>/usr/templates

=item template_file

Holds the selected template_file. This can be used instead of accessing
the template object directly in the init() and header() hooks.

=item default_path

If set, P::A will follow this path for urls that have no destination.
(i.e. http://sample.com/). One of your modules SHOULD set this, because
without it, your P::A is pretty useless.

=item location

Returns the value of $app->request->location.

=item current_module

Returns the name of the module that contains the current page_*() hook
that is being executed.

=item current_hook

Returns the name of the current page_*() hook that is being executed.

=item direct_output

By setting this to a non-zero value from your page_*() hook, your module
indicates to P::A that you do not want to create a Petal object, but
would rather send the response yourself. HTTP headers are still sent,
but the return value of your page_*() hook becomes the output sent to
the client.

=back

=head2 P::A Methods

=over

=item new_accessor

Usage: $app->new_accessor($name);

Creates a new accessor for all modules to use via $app->$name. Use
$app->$name($value) to set the accessor's value.

=item modules

Usage: $app->modules()

Returns an arrayref of the names of all loaded modules. Use in hook() to
request a hook in all modules.

=item hook

Usage: $app->hook($modules, $hook, @args)

Calls a hook in given module(s) passing @args to the hook. Used
internally for all hooks.

=item url

Usage: $app->url($path)

Returns a url to the given path. For example, if your site is
http://sample.com, and you pass it 'module/new', it returns
'http://sample.com/module/new'. Used by the redirect()
method.

=item get

Usage: $app->get($var)

Returns a variable as passed in the url query. In list context, returns
a list of all values with the name $var. In scalar context, returns the
last value in the url with the name $var.

=item post

Usage: $app->post($var)

Same as get(), but for POST variables.

=item param

Usage: $app->param($var)

Combination of get() and post().

=item param2tpl

Usage: $app->param2tpl('var1', 'var2')

Copies values from GET/POST directly to template. Useful for setting
form field values to their passed value.

=item redirect

Usage: return $app->redirect($path)

Redirects the client to the given path. If $path is omitted, client is
redirected to $app->default_path().

=item dbg

Usage: $app->dbg($var)

Prints the value of $var using Data::Dumper to STDERR.

=back

=head2 Modules

To create a module, include it in your modules/ directory and use the
package name: Module::Path::To::Module. For example, if your modules/
directory looks like this:

	modules/
		User.pm
		User/
			Account.pm

Your package name for User.pm is Module::User. Your package name for
Account.pm is Module::User::Account. P::A uses a source filter to map
each module's package name into a unique namespace.

=head1 INSTALLATION

=head2 P::A Installation

	perl Makefile.PL
	make
	make install

=head2 Apache Configuration

	PerlModule Petal::Application
	<Location /myapp/>
		SetHandler perl-script
		PerlHandler Petal::Application
	</Location>

=head2 Web Root Setup
	
Wherever /myapp/ happens to go (on my system, that's /var/www/myapp):

	$ mkdir -p usr/modules
	$ mkdir -p usr/templates

Put your templates in usr/templates. Put your modules in usr/modules.

=head1 CONFIGURING Petal::Application

The configuration of P::A is in the developer's hands.

=head1 EXAMPLES

=head2 usr/templates/site.html

	<html>
		<head>
			<title>Hello, world!</title>
		</head>
		<body>
			<h1>My Site</h1>
			<div petal:replace="structure include: ${content_template}">
				Placeholder
			</div>
		</body>
	</html>

=head2 usr/templates/user/home.html

	<div petal:omit-tag="">
		This is the user home directory.
	</div>

=head2 User.pm

	package Module::User;

	sub new {
		my ($self, $class, $app) = ({}, @_);
		$self->{app} = $self;
		bless $self, $class;

		$app->default_path('user/home'); 
			# set the default path of the site

		return $self;
	}

	sub page_home {
		my ($self, $app) = @_;

		$app->template->param('content_template', 'user/home.html');
	}

=head1 BUGS

None that I know of, yet.

=head1 NOTES

=head1 SEE ALSO

L<Petal> Petal

=head1 AUTHOR

This monstrosity was created by Warren Smith <wsmith@hbslabs.com>
